/* EditMenu.java
 *
 * created: Thu Dec  3 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/EditMenu.java,v 1.45 1999/06/29 13:47:45 kmr Exp $
 **/

package diana.components;

import Diana;
import diana.*;
import diana.sequence.*;
import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.Range;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.Qualifier;
import uk.ac.sanger.pathogens.embl.QualifierVector;
import uk.ac.sanger.pathogens.embl.QualifierParseException;
import uk.ac.sanger.pathogens.embl.InvalidQualifierException;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;

import java.awt.*;
import java.awt.event.*;

/**
 *  A menu with editing commands.
 *
 *  @author Kim Rutherford
 *  @version $Id: EditMenu.java,v 1.45 1999/06/29 13:47:45 kmr Exp $
 **/

public class EditMenu extends SelectionMenu {
  /**
   *  Create a new EditMenu object.
   *  @param frame The Frame that owns this Menu.
   *  @param selection_owner The owner of the selection that the commands in
   *    the menu will operate on.
   *  @param entry_group The EntryGroup object where new features/entries will
   *    be added.
   **/
  public EditMenu (final Frame frame,
                   final SelectionOwner selection_owner,
                   final EntryGroup entry_group) {
    super (frame, "Edit", selection_owner);

    this.entry_group = entry_group;

    edit_feature_item = new MenuItem ("Edit Selected Features",
                                      new MenuShortcut (KeyEvent.VK_E));
    edit_feature_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        editSelectedFeatures ();
      }
    });

    raw_edit_feature_item = new MenuItem ("Raw Edit Selected Features");
    raw_edit_feature_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        rawEditSelectedFeatures ();
      }
    });

    merge_features_item = new MenuItem ("Merge Selected Features");
    merge_features_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        mergeFeatures ();
      }
    });

    duplicate_item = new MenuItem ("Duplicate Selected Features");
    duplicate_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        duplicateFeatures ();
      }
    });

    delete_features_item =
      new MenuItem ("Delete Selected Features",
                    new MenuShortcut (KeyEvent.VK_DELETE));
    delete_features_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        deleteSelectedFeatures ();
      }
    });

    delete_segments_item = new MenuItem ("Delete Selected Exons/Segments");
    delete_segments_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        deleteSelectedSegments ();
      }
    });

    edit_header_item = new MenuItem ("Edit Header Of Default Entry");
    edit_header_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        editHeader ();
      }
    });

    delete_entry_item = new MenuItem ("Delete An Entry ...");
    delete_entry_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        deleteAnEntry ();
      }
    });

    paste_item = new MenuItem ("Paste");
    paste_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        //        pasteSelection (selection);
      }
    });

    duplicate_segments_item = new MenuItem ("Duplicate Selected Segments");
    duplicate_segments_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        duplicateSegments ();
      }
    });

    move_features_item = new MenuItem ("Move Selected Features To ...");
    move_features_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        moveSelectedFeatures ();
      }
    });

    copy_features_to_item = new MenuItem ("Copy Selected Features To ...");
    copy_features_to_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        copySelectedFeaturesTo ();
      }
    });

    fix_stop_codons_item = new MenuItem ("Fix Stop Codons");
    fix_stop_codons_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        fixStopCodons ();
      }
    });

    trim_to_any_item = new MenuItem ("Trim Selected Features To Any");
    trim_to_any_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        trimSelected (true);
      }
    });

    trim_item = new MenuItem ("Trim Selected Features To Met");
    trim_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        trimSelected (false);
      }
    });

    reverse_complement_item = new MenuItem ("Reverse And Complement");
    reverse_complement_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        reverseAndComplement ();
      }
    });

    delete_bases_item = new MenuItem ("Delete Selected Bases");
    delete_bases_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        deleteSelectedBases ();
      }
    });

    add_bases_item = new MenuItem ("Add Bases At Selection");
    add_bases_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        addBases ();
      }
    });

    auto_label_item = new MenuItem ("Automatically Create Gene Names");
    auto_label_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        autoGeneName ();
      }
    });

    add (edit_feature_item);
//    add (raw_edit_feature_item);
    addSeparator ();
    add (edit_header_item);
    addSeparator ();
    add (duplicate_item);
    add (merge_features_item);
    add (delete_features_item);
    add (delete_entry_item);
    addSeparator ();
    add (move_features_item);
    add (copy_features_to_item);
    addSeparator ();
    add (trim_item);
    add (trim_to_any_item);
    addSeparator ();
    add (auto_label_item);
    add (fix_stop_codons_item);
    add (reverse_complement_item);
    add (delete_bases_item);
    add (add_bases_item);

//      add (delete_segments_item);
      //    add (paste_item);
  }

  /**
   *  If more than this number of features are selected than
   *  editSelectedFeatures () etc. will ask for confirmation.
   **/
  private static final int MAXIMUM_SELECTED_FEATURES = 10;

  /**
   *  Open an edit window (FeatureEdit) for each of the selected features.
   *  The edit component will listen for feature change events and update
   *  itself.
   **/
  private void editSelectedFeatures () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really edit all (>" +
                                    MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_edit = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_edit.size () ; ++i) {
      final Feature selection_feature = features_to_edit.elementAt (i);

      new FeatureEdit (selection_feature).show ();
    }
  }

  /**
   *  Open an raw edit (RawFeatureEdit) window for each of the selected
   *  features.  The edit component will listen for feature change events and
   *  update itself.
   **/
  private void rawEditSelectedFeatures () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really edit all (>" +
                                    MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_edit = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_edit.size () ; ++i) {
      final Feature selection_feature = features_to_edit.elementAt (i);

      new RawFeatureEdit (selection_feature);
    }
  }

  /**
   *  Open a EntryHeaderEdit window for the default entry.
   **/
  private void editHeader () {
    final Entry default_entry = entry_group.getDefaultEntry ();

    if (default_entry == null) {
      final String message =
        "there is no default entry";
      new MessageDialog (getParentFrame (), message);
    } else {
      new EntryHeaderEdit (entry_group, default_entry);
    }
  }



  /**
   *  Merge the selected features into one Feature.  If there are selected
   *  segments then the owning Feature of each segment will be the Feature
   *  that is merged.
   **/
  private void mergeFeatures () {
    if (!checkForSelectionFeatures (10, "really merge all (>10) " +
                                    "selected features?")) {
      return;
    }

    final FeatureVector features_to_merge = getSelection ().getAllFeatures ();

    final Feature merge_feature = features_to_merge.elementAt (0);

    // make sure all the features are on the same strand
    for (int i = 1 ; i < features_to_merge.size () ; ++i) {
      final Feature this_feature = features_to_merge.elementAt (i);

      if (this_feature.isForwardFeature () !=
          merge_feature.isForwardFeature ()) {
        new MessageDialog (getParentFrame (),
                           "all the features in a merge must be on the same " +
                           "strand");
        return;
      }

      if (! this_feature.getKey ().equals (merge_feature.getKey ())) {
        new MessageDialog (getParentFrame (),
                           "all the features in a merge must be on the same " +
                           "key");
        return;
      }
    }

    final YesNoDialog dialog =
      new YesNoDialog (getParentFrame (),
                       "Are you sure you want to merge the select " +
                       "features?");

    if (!dialog.getResult ()) {
      return;
    }

    try {
      final Feature new_feature = merge_feature.duplicate ();

      try {

        for (int i = 1 ; i < features_to_merge.size () ; ++i) {
          final Feature this_feature = features_to_merge.elementAt (i);

          final diana.QualifierVector qualifiers =
            this_feature.getQualifiers ();

          for (int qualifier_index = 0 ;
               qualifier_index < qualifiers.size () ;
               ++qualifier_index) {
            final Qualifier this_qualifier =
              qualifiers.elementAt (qualifier_index);

            new_feature.addQualifierValues (this_qualifier);
          }

          final FeatureSegmentVector segments = this_feature.getSegments ();

          for (int segment_index = 0 ;
               segment_index < segments.size () ;
               ++segment_index) {
            final FeatureSegment this_segment =
              segments.elementAt (segment_index);

            final Range this_range = this_segment.getRange ();

            new_feature.addSegment (this_range);
          }
        }

        return;
      } catch (InvalidRelationException e) {
        new_feature.removeFromEntry ();
        throw new Error ("internal error - unexpected exception: " + e);
      }
    } catch (ReadOnlyException e) {
      final String message =
        "the destination feature is read-only or is in a read-only entry";
      new MessageDialog (getParentFrame (), message);
      return;
    }
  }

  /**
   *  Duplicate the selected Feature objects.  If there are selected segments
   *  then the owning Feature of each segment will be duplicated.
   **/
  private void duplicateFeatures () {
    if (!checkForSelectionFeatures (100,
                                    "really duplicate all (>100) " +
                                    "selected features?")) {
      return;
    }

    final FeatureVector features_to_duplicate =
      getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_duplicate.size () ; ++i) {
      final Feature this_feature = features_to_duplicate.elementAt (i);

      try {
        this_feature.duplicate ();
      } catch (ReadOnlyException e) {
        final String message =
          "one of the selected features (" + this_feature.getIDString () +
          ")  is read only - cannot continue";
        new MessageDialog (getParentFrame (), message);
        return;
      }
    }

//    Diana.getClipboard ().setClip (getSelection ());
  }

  /**
   *  Duplicate the selected Feature objects.
   **/
  private void duplicateSegments () {
    if (!checkForSelectionFeatures ()) {
      return;
    }

    final FeatureSegmentVector segments_to_duplicate =
      getSelection ().getAllSegments ();

    for (int i = 0 ; i < segments_to_duplicate.size () ; ++i) {
      final FeatureSegment this_segment = segments_to_duplicate.elementAt (i);
      this_segment.duplicate ();
    }

//    Diana.getClipboard ().setClip (getSelection ());
  }

  /**
   *  Delete the selected features.
   **/
  private void deleteSelectedFeatures () {
    final FeatureVector features_to_delete = getSelection ().getAllFeatures ();

    // 0 means always popup a YesNoDialog
    if (!checkForSelectionFeatures (0, "really delete " +
                                    (features_to_delete.size () == 1 ?
                                     "the selected feature?" :
                                     features_to_delete.size () +
                                     " features?"))) {
      return;
    }

    for (int i = 0 ; i < features_to_delete.size () ; ++i) {
      final Feature selection_feature = features_to_delete.elementAt (i);

      try {
        selection_feature.removeFromEntry ();
      } catch (ReadOnlyException e) {
        final String message =
          "one of the selected features (" +
          selection_feature.getIDString () +
          ") is in a read only entry - cannot continue";
        new MessageDialog (getParentFrame (), message);
        return;
      }

      getSelection ().remove (selection_feature);
    }
  }

  /**
   *  Delete the selected feature segments.
   **/
  private void deleteSelectedSegments () {
    final FeatureSegmentVector segments_to_delete =
      getSelection ().getSelectedSegments ();

    for (int i = 0 ; i < segments_to_delete.size () ; ++i) {
      final FeatureSegment selection_segment =
        segments_to_delete.elementAt (i);

      selection_segment.removeFromFeature ();
    }
  }

  /**
   *  Popup a EntryChooser component and then delete the Entry that the
   *  user chooses.
   **/
  private void deleteAnEntry () {
    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose an entry to delete:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          entry_group.remove (event.getChoosenEntry ());
          event.getChoosenEntry ().dispose ();
          entry_chooser.dispose ();
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Popup a EntryChooser component and then move the selected features to
   *  Entry that the user chooses.
   **/
  private void moveSelectedFeatures () {
    if (!checkForSelectionFeatures ()) {
      return;
    }

    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose a destination entry:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          moveFeaturesHelper (getSelection ().getAllFeatures (),
                              event.getChoosenEntry ());
          entry_chooser.dispose ();
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Move the given features to the given destination entry.
   **/
  private void moveFeaturesHelper (final FeatureVector features,
                                   final Entry destination_entry) {
    for (int i = 0 ; i < features.size () ; ++i) {
      final Feature this_feature = features.elementAt (i);
      try {
        this_feature.moveTo (destination_entry);
      } catch (ReadOnlyException e) {
        final String message =
          "one of the selected features (" +
          this_feature.getIDString () +
          ") is in a read only entry - cannot continue";
        new MessageDialog (getParentFrame (), message);
        return;
      }
    }
  }

  /**
   *  Popup a EntryChooser component and then move the selected features to
   *  Entry that the user chooses.
   **/
  private void copySelectedFeaturesTo () {
    if (!checkForSelectionFeatures (100,
                                    "really copy all (>100) " +
                                    "selected features?")) {
      return;
    }

    final EntryChooser entry_chooser =
      new EntryChooser (entry_group, "choose a destination entry:");

    final EntryChooserListener new_listener =
      new EntryChooserListener () {
        public void entryChoosen (EntryChooserEvent event) {
          try {
            final FeatureVector features_to_copy =
              getSelection ().getAllFeatures ();

            for (int i = 0 ; i < features_to_copy.size () ; ++i) {
              final Feature this_feature = features_to_copy.elementAt (i);
              this_feature.copyTo (event.getChoosenEntry ());
            }
          } catch (ReadOnlyException e) {
            final String message =
              "the destination entry is read only - cannot continue";
            new MessageDialog (getParentFrame (), message);
          } finally {
            entry_chooser.dispose ();
          }
        }
      };

    entry_chooser.addEntryChooserListener (new_listener);

    entry_chooser.setVisible (true);
  }

  /**
   *  Calls fixStopCodon () on each selected feature.
   **/
  private void fixStopCodons () {
    if (!checkForSelectionFeatures ()) {
      return;
    }

    final FeatureVector features_to_fix = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_fix.size () ; ++i) {
      final Feature selection_feature = features_to_fix.elementAt (i);

      if (!selection_feature.fixStopCodon ()) {
        // select the bad feature
        getSelection ().set (selection_feature);

        final String message =
          "could not fix the stop codon for: " +
          selection_feature.getIDString ();

        new MessageDialog (getParentFrame (), message);

        return;
      }
    }
  }

  /**
   *  Calls Feature.trimStart () on all selected Feature objects.
   *  @param trim_to_any If true then the features will be trimmed to ATG, GTG
   *    or TTG, otherwise the features will be trimmed to ATG only.
   **/
  private void trimSelected (final boolean trim_to_any) {
    if (!checkForSelectionFeatures ()) {
      return;
    }

    final FeatureVector features_to_trim = getSelection ().getAllFeatures ();

    // this will contain the features that could not be trimmed
    final FeatureVector failed_features = new FeatureVector ();

    for (int i = 0 ; i < features_to_trim.size () ; ++i) {
      final Feature selection_feature = features_to_trim.elementAt (i);

      if (!selection_feature.trimStart (trim_to_any)) {
        failed_features.add (selection_feature);
      }
    }

    if (failed_features.size () > 0) {
      // select the bad feature
      getSelection ().set (failed_features);

      final String message =
        "some of the features could not be trimmed (they are now selected)";

      new MessageDialog (getParentFrame (), message);
    }
  }

  /**
   *  Reverse and complement the sequence and all the features in all Entry
   *  objects.
   **/
  private void reverseAndComplement () {
    final YesNoDialog dialog =
      new YesNoDialog (getParentFrame (),
                       "Are you sure you want to reverse complement all " +
                       "entries?");

    if (!dialog.getResult ()) {
      return;
    }

    try {
      entry_group.reverseComplement ();
    } catch (ReadOnlyException e) {
      final String message =
        "one of the entries is read only - cannot continue";
      new MessageDialog (getParentFrame (), message);
      return;
    }
  }

  /**
   *  Delete the selected bases after asking the user for confimation.
   **/
  private void deleteSelectedBases () {
    if (!checkForSelectionRange ()) {
      return;
    }

    final MarkerRange range =
      getSelectionOwner ().getSelection ().getMarkerRange ();

    final FeatureVector features_in_range;

    try {
      features_in_range =
        entry_group.getFeaturesInRange (range.getRawRange ());
    } catch (OutOfRangeException e) {
      throw new Error ("internal error - unexpected exception: " + e);
    }

    if (features_in_range.size () == 0) {
      final YesNoDialog dialog =
        new YesNoDialog (getParentFrame (),
                         "Are you sure you want to delete the " +
                         "selected bases?");

      if (dialog.getResult ()) {
        // deleteRange () is static and uses the strand reference from the
        // MarkerRange to decide which Strand to edit
        Strand.deleteRange (range);
        getSelectionOwner ().getSelection ().setMarkerRange (null);
      }
    } else {
      new MessageDialog (getParentFrame (),
                         "cannot delete - there are features within the " +
                         "selected range");
    }
  }

  /**
   *  Ask the user for some bases and then add them at the start of the
   *  selected range.
   **/
  private void addBases () {
    if (!checkForSelectionRange ()) {
      return;
    }

    final MarkerRange range =
      getSelectionOwner ().getSelection ().getMarkerRange ();

    final TextRequester text_requester =
      new TextRequester ("enter the bases to insert before base " +
                         range.getStart ().getPosition () +
                         (range.isForwardMarker () ?
                          " on the forward strand" :
                          " on the reverse strand") + ":",
                         18, "");

    text_requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        if (event.getType () == TextRequesterEvent.CANCEL) {
          return;
        }

        final String bases_string = event.getRequesterText ().trim ();

        if (bases_string.length () == 0) {
          new MessageDialog (getParentFrame (), "no bases inserted");
          return;
        }

        // addBases () is static and uses the strand reference from the start
        // Marker to decide which Strand to edit
        Strand.addBases (range.getStart (), bases_string);
      }
    });

    text_requester.show ();
  }


  /**
   *  Automatically create systematic gene names for all CDS features in the
   *  active entries.
   **/
  private void autoGeneName () {
    // first make sure that none of the CDS features have gene names
    final FeatureEnumeration test_enumerator = entry_group.features ();

    while (test_enumerator.hasMoreFeatures ()) {
      final Feature this_feature = test_enumerator.nextFeature ();

      final Key key = this_feature.getKey ();

      if (key.equals ("CDS") &&
          this_feature.getGeneName () != null &&
          ! this_feature.getGeneName ().equals ("none")) {
        getSelection ().set (this_feature);

        new MessageDialog (getParentFrame (),
                           "cannot continue - this feature has a gene name: " +
                           this_feature.getGeneName ());
        return;
      }
    }

    final TextRequester text_requester =
      new TextRequester ("enter the start characters of the new gene names:",
                         18, "");

    text_requester.addTextRequesterListener (new TextRequesterListener () {
      public void actionPerformed (final TextRequesterEvent event) {
        if (event.getType () == TextRequesterEvent.CANCEL) {
          return;
        }

        final String name_start_string = event.getRequesterText ().trim ();

        if (name_start_string.length () == 0) {
          new MessageDialog (getParentFrame (), "no name given");
          return;
        }

        int current_number = 1;

        // now name the CDS features
        final FeatureEnumeration feature_enumerator = entry_group.features ();

        while (feature_enumerator.hasMoreFeatures ()) {
          final Feature this_feature = feature_enumerator.nextFeature ();

          final Key key = this_feature.getKey ();

          if (key.equals ("CDS")) {
            final String number_string;

            if (current_number < 10) {
              number_string = "0" + current_number;
            } else {
              number_string = String.valueOf (current_number);
            }

            try {
              final Qualifier new_qualifier =
                new Qualifier ("gene", name_start_string + "." +
                               number_string +
                               (this_feature.isForwardFeature () ? "" : "c"));

              this_feature.setQualifier (new_qualifier);
            } catch (QualifierParseException e) {
              throw new Error ("internal error - unexpected exception: " + e);
            } catch (InvalidQualifierException e) {
              throw new Error ("internal error - unexpected exception: " + e);
            } catch (InvalidRelationException e) {
              throw new Error ("internal error - unexpected exception: " + e);
            } catch (ReadOnlyException e) {
              new MessageDialog (getParentFrame (),
                                 "one of the feature is in a read-only " +
                                 "entry - cannot continue");
              return;
            }

            ++current_number;
          }
        }
      }
    });

    text_requester.show ();
  }


  /**
   *  The EntryGroup object that was passed to the constructor.
   **/
  final EntryGroup entry_group;

  private MenuItem edit_feature_item = null;
  private MenuItem raw_edit_feature_item = null;
  private MenuItem duplicate_item = null;
  private MenuItem merge_features_item = null;
  private MenuItem duplicate_segments_item = null;
  private MenuItem delete_features_item = null;
  private MenuItem edit_header_item = null;
  private MenuItem delete_segments_item = null;
  private MenuItem delete_entry_item = null;
  private MenuItem paste_item = null;
  private MenuItem move_features_item = null;
  private MenuItem copy_features_to_item = null;
  private MenuItem trim_item = null;
  private MenuItem trim_to_any_item = null;
  private MenuItem auto_label_item = null;
  private MenuItem fix_stop_codons_item = null;
  private MenuItem reverse_complement_item = null;
  private MenuItem delete_bases_item = null;
  private MenuItem add_bases_item = null;
}
